/* $Id: CbcBranchActual.hpp 2465 2019-01-03 19:26:52Z unxusr $ */
// Copyright (C) 2002, International Business Machines
// Corporation and others.  All Rights Reserved.
// This code is licensed under the terms of the Eclipse Public License (EPL).

#ifndef CbcBranchActual_H
#define CbcBranchActual_H

#include "CbcBranchBase.hpp"
#include "CoinPackedMatrix.hpp"
#include "CbcClique.hpp"
#include "CbcSOS.hpp"
#include "CbcSimpleInteger.hpp"
#include "CbcNWay.hpp"
#include "CbcSimpleIntegerPseudoCost.hpp"
#include "CbcBranchDefaultDecision.hpp"
#include "CbcFollowOn.hpp"
#include "CbcFixVariable.hpp"
#include "CbcDummyBranchingObject.hpp"
#include "CbcGeneral.hpp"
#include "CbcGeneralDepth.hpp"
#include "CbcSubProblem.hpp"
#endif

/* vi: softtabstop=2 shiftwidth=2 expandtab tabstop=2
*/
